/*--------------------------------------------------------------------------
**      Critical.c                 Twin           From:  Twin/scheduler
**
**              
**------------------------- < License Information > ------------------------
**
**      This file was originally a part of Willows TWIN.  Willows
**  TWIN was released under a Library GPL (LGPL).  This permits
**  redistribution of this source code, provided that the full
**  TWIN license is in effect, and provided that all modifications
**  to this source code are made publicly available.
**  Please refer to Willows software (www.willows.com) or
**  LICENSE for full information.
**  
**      Under Twine, this file is also protected by an LGPL.  Please
**  see LICENSE for full details on this license.
**  
**
**      
**------------------------ < File Content Description > --------------------
**
**  Module:	scheduler/Critical.c
**
**  Description:
**      
**
**  Functions defined:
**    
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/scheduler/Critical.c,v 1.7 1999/12/08 19:27:03 robf Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
**              E X T E R N A L     I N C L U D E S 
**--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
**              I N T E R N A L     I N C L U D E S
**--------------------------------------------------------------------------*/
#include "windows.h"
#include "stdio.h"
#include "Log.h"
#include "kerndef.h"
#include "Resources.h"
#include "Module.h"
#include "osdll.h"
#include "strings.h"
#include "mtlock.h"

/*******************************************************
 *
 * critical sections
 *
 ******************************************************/

/* the system locks */
/* see: mtlock.h    */
CRITICAL_SECTION  _locktable[_NMTLOCKS];

/********************************/
/* debug info wrt/ system locks */
/********************************/
typedef struct {
	int			init;		// lock initialized
	int			id;		// id of lock, safeguard
	char		       *name;		// name of lock
	char		       *file;		// where currently locked
	int			line;		// line# where locked
	int			count;		// types locked/unlocked
	int			active;		// types locked/unlocked
	long			thread;
} LOCKINFO;

LOCKINFO _lockinfo[_NMTLOCKS] = {
	{ 0,_MTLOCK_TABLE,"Locktable",0,0,0,0,0},
	{ 0,_MTLOCK_WIN16,"Win16 Mutex",0,0,0,0,0},
	{ 0,_MTLOCK_VIDEO,"Video",0,0,0,0,0},
	{ 0,_MTLOCK_QUEUE,"Message Queue",0,0,0,0,0},
	{ 0,_MTLOCK_THREAD,"Thread",0,0,0,0,0},
	{ 0,_MTLOCK_WINDOW,"Window",0,0,0,0,0},
	{ 0,_MTLOCK_PROCESS,"Process",0,0,0,0,0},
	{ 0,_MTLOCK_MEMORY,"Memory",0,0,0,0,0},
};

void
twin_assert(int val,char *p,int l)
{
	if(val == 0) {
		logstr(LF_CONSOLE,"TWIN_ASSERT: %d failed %s:%d\n",
			val,p,l);
	}
	// break on assert fail
	// abort on assert fail
}

/*
 * one shot, initialize system critical sections
 */
void 
TWIN_InitializeCriticalSections()
{
	InitializeCriticalSection(&_locktable[_MTLOCK_TABLE]);
	_lockinfo[_MTLOCK_TABLE].init++;

	InitializeCriticalSection(&_locktable[_MTLOCK_WIN16]);
	_lockinfo[_MTLOCK_WIN16].init++;

	InitializeCriticalSection(&_locktable[_MTLOCK_VIDEO]);
	_lockinfo[_MTLOCK_VIDEO].init++;
}

void
TWIN_DCS()
{
}

void
TWIN_EnterSection(int id,char *file,int line)
{
	EnterCriticalSection(&_locktable[id]);
}

void
TWIN_LeaveSection(int id)
{
	LeaveCriticalSection(&_locktable[id]);
}

/********************************************************/
/*							*/
/* critical section API's 				*/
/*							*/
/********************************************************/

void WINAPI 
InitializeCriticalSection( CRITICAL_SECTION *crit )
{
	crit->LockCount = -1;
    	crit->RecursionCount = 0;
    	crit->OwningThread   = 0;
    	crit->LockSemaphore  = 0;
	crit->SpinCount      = 0;
}

void WINAPI 
DeleteCriticalSection( CRITICAL_SECTION *crit )
{
	crit->OwningThread = 0;
	crit->LockCount    = 0;
	crit->SpinCount    = 0;
}

#define THISTHREAD	getpid()

static WAITNODE *waitlisthead;
static DWORD     waitlistlock;

WAITNODE *
TWIN_GetWaitNode()
{
	WAITNODE *waitnode;

	acquire(&waitlistlock);
	waitnode = waitlisthead;
	if(waitnode)
		waitlisthead = waitlisthead->wn_next;
	release(&waitlistlock);

	return waitnode?waitnode:
		(WAITNODE *) WinMalloc(sizeof(WAITNODE));
}
void
TWIN_FreeWaitNode(WAITNODE *waitnode)
{
	acquire(&waitlistlock);
	waitnode->wn_next = waitlisthead;
	waitlisthead = waitnode;
	release(&waitlistlock);
}

/************************************************************************
 * 
 * Discussion:
 * ReleaseCriticalSection()
 * if we are called, then we know that one or more threads
 * have been queued, and are or will be waiting for a signal
 *
 * we also know the signal is blocked on the thread that will be 
 * waiting, up through and including the time they actually go to
 * sleep, in sigsuspend
 *
 * we will only call releasecriticalsection, when it is guaranteed
 * that another thread is actually queued, ie. they incremented
 * the lockcount, while they had the spinlock, and they queued 
 * themselves, all in one go...
 */

static void
ReleaseCriticalSection( CRITICAL_SECTION *crit )
{
        THDB *pThread = 0;
	WAITNODE *waitlist;

	/* find a thread that is waiting  */
	/* spinlock to get semaphore list */
	/* yield to other threads         */
	acquire(&crit->SpinCount);

	/* get first element on list */
	/* remove from queue         */
	if((waitlist = (WAITNODE *) crit->LockSemaphore))
		crit->LockSemaphore = (DWORD) waitlist->wn_next;

	/* release the spinlock */
	/* we only hold it to lock the queue */

	release(&crit->SpinCount);

	/* find waitlist element, free it */
	/* this MUST be true 		  */
	if(waitlist) {
		pThread = (THDB *) waitlist->wn_thread;
		TWIN_FreeWaitNode(waitlist);
	} 

	/* this is the thread that has incremented */
	/* the lockcount, this MUST be true        */
	if(pThread) {
		osdll_thread_ops(pThread,OSDLL_THREAD_RESTART);
	}

	return;
}

static void
WaitCriticalSection( CRITICAL_SECTION *crit )
{
        THDB *pThread = osdll_thread_gettdb();
	WAITNODE *waitnode;
	WAITNODE *waitlist;

	/* we are still holding the original spinlock */
	waitnode = TWIN_GetWaitNode();
	waitnode->wn_thread = pThread;
	waitnode->wn_wait   = 0;
	waitnode->wn_object = 0;
	waitnode->wn_next   = 0;

	/* this is why we are holding the spinlock */
	/* to queue ourselves */
    	if((waitlist = (WAITNODE *) crit->LockSemaphore) == 0)
		crit->LockSemaphore = (DWORD) waitnode;
	else {
		while(waitlist->wn_next)
			waitlist = waitlist->wn_next;
		waitlist->wn_next = waitnode;
	}

	/* now we are queued */
	/* we will either be asleep, or on our way */
	/* on the next call */
	release(&crit->SpinCount);

	osdll_thread_suspend();

	/* this matches the acquire that we initially */
	/* acquired in EnterCriticalSection, but      */
	/* released above before waiting              */

	acquire(&crit->SpinCount);
}

void WINAPI 
EnterCriticalSection( CRITICAL_SECTION *crit )
{
//	APISTR((LF_APICALL, "EnterCriticalSection(%p)\n",crit));

	/* acquire the spinlock */
	acquire(&crit->SpinCount);

	/* increment the lockcount */
    	if (InterlockedIncrement( &crit->LockCount )) {

		/* we own it, increment recursion count */
		if (crit->OwningThread == THISTHREAD) {
		    release(&crit->SpinCount);

		    crit->RecursionCount++;
		    //APISTR((LF_APIRET, "EnterCriticalSection returns void\n"));
		    return;
		}
		
		/* hold long enough to queue us */
		WaitCriticalSection(crit);
	}
	release(&crit->SpinCount);

    	crit->OwningThread   = THISTHREAD;
    	crit->RecursionCount = 1;

//	APISTR((LF_APIRET, "EnterCriticalSection returns void\n"));
}

#define SPIN_TRYWAIT	10

/* no spinlocks required here, we will NOT be queued,
 * and not block.  if we cannot outright acquire the
 * lock, we return failure.
 *
 * the key here is simply the lockcount, it is either
 * free or owned, if owned, we will yield, and try
 * again.
 * 
 * if another thread EnterCriticalSection, and acquires 
 * the spinlock if we were first, they will be queued, 
 * and we will wake them when we LeaveCritical section
 *
 * if another thread EnterCriticalSection, and acquires
 * the spinlock, and beat us to the lockcount, then we
 * will not block, but return failed to acquire he 
 * section
 */
BOOL WINAPI 
TryEnterCriticalSection( CRITICAL_SECTION *crit )
{
	int spincount = 0;    

    	while (InterlockedIncrement( &crit->LockCount )) {
		spincount++;

        	if (crit->OwningThread == THISTHREAD) {
            		crit->RecursionCount++;
            		return TRUE;
        	}

		/* if someone owns it still, we fail */
        	if(InterlockedDecrement( &crit->LockCount ))
        		return FALSE;

		/* if may be available, try TRYWAIT times */
		if(spincount > SPIN_TRYWAIT)
			return FALSE;
    	}

	/* we acquired it */
    	crit->OwningThread   = THISTHREAD;
    	crit->RecursionCount = 1;
    	return TRUE;
}

/*
 * the key to someone waiting is the lockcount
 * if set when we leave, someone else wants it
 * try to wake them up
 */
void WINAPI 
LeaveCriticalSection( CRITICAL_SECTION *crit )
{
//	APISTR((LF_APICALL, "LeaveCriticalSection(%p)\n",crit));
    	if (crit->OwningThread != THISTHREAD) {
		//APISTR((LF_APIRET, "LeaveCriticalSection (unowned section) returns void\n"));
		return;
	}

    	if (--crit->RecursionCount) {
        	InterlockedDecrement( &crit->LockCount );
		//APISTR((LF_APIRET, "LeaveCriticalSection returns void\n"));
        	return;
    	}

	/* from here on, someone can try, or enter this */
	/* section. */
    	crit->OwningThread = 0;

    	if (InterlockedDecrement( &crit->LockCount ) >= 0) {
		ReleaseCriticalSection(crit);
    	}
//	APISTR((LF_APIRET, "LeaveCriticalSection returns void\n"));
}
 
void WINAPI
ReinitializeCriticalSection( CRITICAL_SECTION *crit )
{
    	DeleteCriticalSection( crit );
    	InitializeCriticalSection( crit );
}

typedef struct _tagMUTEX {
	struct _tagMUTEX *next;
	SECURITY_ATTRIBUTES security;
	DWORD             access;
	BOOL		  inherit;
	char		 *name;	
	HANDLE		  hMutex;
} MUTEX;

/* (WIN32) Mutex *********************************************************** */

#ifdef TWIN_BIN32

HANDLE
CreateMutexA(LPSECURITY_ATTRIBUTES *lpsec,BOOL bInherit,LPCSTR name)
{
	return 0;
}

HANDLE
OpenMutexA(DWORD access,BOOL binherit,LPCSTR lpcstr)
{
	return (HANDLE) 0;
}
#endif

HANDLE	WINAPI
CreateMutex(LPSECURITY_ATTRIBUTES lpMutexAttributes,
	BOOL bInitialOwner,
	LPCSTR lpName)
{
	APISTR((LF_APISTUB, "CreateMutex: (API) %s\n", lpName));
	return ((HANDLE) 0);
}

HANDLE	WINAPI
OpenMutex(DWORD dwDesiredAccess,
	BOOL bInheritHandle,
	LPCSTR lpName)
{
	APISTR((LF_APISTUB, "OpenMutex: (API) %s\n", lpName));
	return ((HANDLE) 0);
}

BOOL	WINAPI
ReleaseMutex(HANDLE hMutex)
{
	APISTR((LF_APISTUB, "ReleaseMutex: (API) hMutex %x\n", hMutex));
	return (TRUE);
}

BOOL
DisableThreadLibraryCalls(HMODULE hModule)
{
	return TRUE;
}

/*------------------------- < Full Revision History > ----------------------
** $Log: Critical.c,v $
** Revision 1.7  1999/12/08 19:27:03  robf
** disable logging of enter/exit critical sections
**
** Revision 1.5  1999/07/13 20:17:37  mwalsh
** Updated comments to use standard header and footer.
**
**-------------------------------------------------------------------------*/
