/*--------------------------------------------------------------------------
**      ObjEngine.c                 Twin           From:  Twin/windows
**
**              
**------------------------- < License Information > ------------------------
**
**      This file was originally a part of Willows TWIN.  Willows
**  TWIN was released under a Library GPL (LGPL).  This permits
**  redistribution of this source code, provided that the full
**  TWIN license is in effect, and provided that all modifications
**  to this source code are made publicly available.
**  Please refer to Willows software (www.willows.com) or
**  LICENSE for full information.
**  
**      Under Twine, this file is also protected by an LGPL.  Please
**  see LICENSE for full details on this license.
**  
**
**      Copyright 1997 Willows Software, Inc.
**------------------------ < File Content Description > --------------------
**
**  Module:	windows/ObjEngine.c
**
**  Description:
**      
**
**  Functions defined:
**    
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/windows/ObjEngine.c,v 1.9 1999/12/14 17:49:17 robf Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/



/*
*
*
*	An object has a OBJHEAD, which contains:
*		hObj		the HANDLE for this object
*		wObjSignature	a double word of type
*				eg. MAKESIGNATURE(OT_GDI,OT_BRUSH)
*
*	The following fields are also present, and are defined by the
*	object owner.
*		wFlags		object flags, tentatively
*				selected,
*				deleted
*		wRefCount	usage count how many people are using it
*				(ie. is it currently locked, by how many)
*		lpObjNext	linked list to additional objects
*
********************************************************************/

/*----------------------------------------------------------------------------
**              E X T E R N A L     I N C L U D E S 
**--------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>

/*----------------------------------------------------------------------------
**              I N T E R N A L     I N C L U D E S
**--------------------------------------------------------------------------*/
#include "windows.h"
#include "ObjEngine.h"
#include "GdiObjects.h"
#include "Kernel.h"
#include "WinDefs.h"
#include "GdiDC.h"
#include "GdiText.h"
#include "Listbox.h"
#include "Scroll.h"
#include "Module.h"
#include "Log.h"
#include "compat.h"
#include "WinMalloc.h"
#include "thread.h"
#include "mtlock.h"

#ifdef DEBUG
extern void debuggerbreak();
#endif

void DEBUG_DumpHandle();

#ifdef DEBUG_OBJECTS
static char *cmdnames[] = {
	"HM_INIT     "  ,
	"HM_GETHND   "  ,
	"HM_GETOBJ   "  ,
	"HM_FREEHND  "  ,
	"HM_LOCKHND  "	,
	"HM_UNLOCKHND"	,
	"HM_DUMPTABLE"	,
	"HM_CHECKHND "
};
#endif

int GetHandleErr(int,HANDLE,WORD,WORD);

/*
 *	GetHandleErr
 *	used primarily for debug purposes
 *	set breakpoint at GetHandleErr, and check errcode
 *	fatal errors will not return
 */

/* set this to a handle, we will break if we see it */
static HANDLE BreakObject = 0;
static int    BreakFunc   = 0;

/* this is the handle table */
/* indexed by the HIWORD of a handle,  */
/* then the LOWORD of the handle       */
static LPOBJHEAD **lplpArray = 0;

/* lock against getting/releasing objects */
static DWORD    FreeLock = 0;
static DWORD	lfobj	 = LF_CONSOLE;

LPOBJHEAD
newhandle(int Size)
{
	LPOBJHEAD  *lpObjList;
	LPOBJHEAD  lpObj;
	int IndexHi,IndexLo;
	static int freelist = 1;

	acquire(&FreeLock);
    	if (lplpArray == 0) {
		/* allocate the object table, HIWORD index */
		lplpArray = (LPOBJHEAD **)WINMALLOC(sizeof(LPOBJHEAD) * 256,5,0);
		memset((LPSTR)lplpArray, '\0', sizeof(LPOBJHEAD) * 256);
    	}

	for(;;) {
		/* don't use slot 0 */
		for(IndexHi = 1;IndexHi < 256;IndexHi++) {
			lpObjList = lplpArray[IndexHi];
			if(lpObjList) {
				for(IndexLo=0;IndexLo < 256;IndexLo++) {
					lpObj = lpObjList[IndexLo];
					if(lpObj)
						continue;
				
	    				lpObjList[IndexLo] = WINMALLOC(Size,0,0);
					lpObj = lpObjList[IndexLo];
            				memset(lpObj, 0, Size);
						
					/* set its HANDLE */
	    				lpObj->hObj =  (IndexHi<<8)|(IndexLo);
					/* set its refcnt */
	    				lpObj->wRefCount = 1;

					release(&FreeLock);
					return lpObj;
				}
			}
		}
	
		if(freelist == 256) {
			logstr(lfobj,"OBJMGR: handle table full\n");
			debuggerbreak();
			return 0;
		}
    		lplpArray[freelist] = (LPOBJHEAD *)WINMALLOC(sizeof(LPOBJHEAD) * 256,5,0);
    		memset((LPSTR)lplpArray[freelist], '\0', sizeof(LPOBJHEAD)*256);
		freelist++;
	}
}

LPOBJHEAD
freehandle(DWORD hObj)
{
	LPOBJHEAD lpObj;
	int IndexHi = HIBYTE((WORD)(hObj));
	int IndexLo = LOBYTE((WORD)(hObj));

	acquire(&FreeLock);
	if (!lplpArray[IndexHi]) {
		return (LPOBJHEAD)GetHandleErr(5,hObj,0,0);
	}

	//logstr(lfobj,"OBJMGR: freeing %x\n",hObj);
	lpObj = lplpArray[IndexHi][IndexLo];

	if(lpObj)
		WinFree(lpObj);
    	lplpArray[IndexHi][IndexLo] = 0;

	release(&FreeLock);
	return 0;
}

LPOBJHEAD
getptr(DWORD hObj,BOOL bExists)
{
	LPOBJHEAD lpObj;
	int IndexHi = HIBYTE((WORD)(hObj));
	int IndexLo = LOBYTE((WORD)(hObj));

	if (!lplpArray[IndexHi])
		return (LPOBJHEAD)GetHandleErr(5,hObj,0,0);

	lpObj = lplpArray[IndexHi][IndexLo];
	if(bExists && lpObj == 0) {
		return (LPOBJHEAD)GetHandleErr(6,hObj,0,0);
	}
	return lpObj;
}

/*****************************************************************/

char *
sigof(int Func,HANDLE h,int x)
{
	static char signature[3];

	signature[0] = (x >> 8) & 0xff;
	signature[1] = (x     ) & 0xff;
	signature[2] = 0;

	if(BreakObject == h || BreakFunc == Func) {
		//logstr(lfobj,"OBJMGR: break %x,%x,%s)\n",
		//	Func,h,signature);
		debuggerbreak();
	}

	return signature;
}

char *
sigstr(int sig)
{
	char *lpString = "BOGUS";

	switch (sig) {
		case MLDC:
			lpString = "DC    ";
			break;
		case MLICON:
			lpString = "ICON  ";
			break;
		case MLBITMAP:
			lpString = "BITMAP";
			break;
		case MLHANDLE:
			lpString = "HANDLE";
			break;
		case MLREGION:
			lpString = "REGION";
			break;
		case MLMODULE:
			lpString = "MODULE";
			break;
		case MLTASK:
			lpString = "TASK  ";
			break;
		case MLBRUSH:
			lpString = "BRUSH ";
			break;
		case MLPEN:
			lpString = "PEN   ";
			break;
		case MLFONT:
			lpString = "FONT  ";
			break;
		case MLPALETTE:
			lpString = "PALETE";
			break;
		case MLWIN:
			lpString = "WINDOW";
			break;
		case MLPOS:
			lpString = "WINPOS";
			break;
		case MLLBOX:
			lpString = "LBINFO";
			break;
		case MLSBAR:
			lpString = "SBINFO";
			break;
		case MLMETAFILE:
			lpString = "MFILE ";
			break;
		case MLCURSOR:
			lpString = "CURSOR";
			break;
	}
	return lpString;
}

int
GetHandleErr(int errcode,HANDLE hObj,WORD wSignature,WORD arg)
{
	int	fatal;
	int  	errflag;
	int	warnflag;

	static char	*fmterr = "***ERROR*** Object Manager:";

	/* not necessarily fatal */
	fatal = 0;

	/* what type of control do we want */

	errflag = LF_ERROR;
	warnflag = LF_WARNING;

	switch(errcode) {
	   case 1:
		fatal = -1;
		ERRSTR((errflag, "%s %s\n",fmterr,"Internal Error"));
		break;

	   case 2:
		ERRSTR((errflag, "%s %s\n",fmterr,"create w/o signature"));
		break;
	   case 3:
	        ERRSTR((errflag, "%s %s %d\n",fmterr,"create w/no object size",wSignature));
		break;
	   case 4:	/* bogus handle 	*/
	        LOGSTR((LF_WARNING, "%s %s\n",fmterr,"get w/ null handle"));
		break;
	   case 5:	/* bad table index    	*/
	        LOGSTR((LF_WARNING, "%s %s %x\n",fmterr,"bad handle",hObj));
		break;

	   case 6:	/* bad table offset	*/
	        LOGSTR((LF_WARNING, "%s %s %x\n",fmterr,"bad offset",hObj));
		break;
	   case 8:	/* bad category match   */
	        LOGSTR((LF_WARNING, "%s %s %x\n",fmterr,"bad category",hObj));
		break;
	   case 9:	/* bad type match   	*/
	        LOGSTR((LF_WARNING, "%s %s %x\n",fmterr,"bad type",hObj));
		break;

	}
#ifdef DEBUG
	if (GetConfigOption(WD_HANDLERR)) {
	        ERRSTR((errflag,fmterr,"Internal Error","Bad Function",hObj));
		debuggerbreak();
	}
#endif

	if (fatal) {

		/* fatal exit, does not return */
		FatalAppExit(fatal,"Object Manager Fatal Error\n");	
	}

	return 0;
}

BOOL TWIN_DeleteObject(LPOBJHEAD);
BOOL TWIN_DeleteHandle(LPHANDLEINFO );
BOOL TWIN_DeleteDC(HDC32 );
BOOL TWIN_DeleteWindow(HWND32);
BOOL TWIN_DeleteLBox(LPLISTBOXINFO);
BOOL TWIN_DeleteIcon(LPTWIN_ICONINFO);
BOOL TWIN_DeleteCursor(LPCURSORINFO);

static int 
deleteobject( HANDLE hObj, LPOBJHEAD lpTemp,WORD wSignature)
{
    	WORD Type, Category;

    	Type = LOBYTE(wSignature);
    	Category = HIBYTE(wSignature);

	/* do not free if category or type does not match */
	if (Category && (HIBYTE(lpTemp->wObjSignature) != Category)) {
	    return GetHandleErr(8,hObj,wSignature,Category);
	}

	if (Type && (LOBYTE(lpTemp->wObjSignature) != Type)) {
	    return GetHandleErr(9,hObj,wSignature,Type);
	}

	/* do not free if locked */
	if(lpTemp->wRefCount) {

		lpTemp->wFlags |= OE_DELETE;
		//logstr(lfobj,"OBJMGR DELAY %x S=%s #=%d F=%x\n",
		//	hObj, sigstr(lpTemp->wObjSignature),
		//	lpTemp->wRefCount,
		//	lpTemp->wFlags);

		return 0;
	}
	
	/* remove the object */
	switch (lpTemp->wObjSignature) {
	    case MLWIN:
		TWIN_DeleteWindow((HWND32)lpTemp);
		break;

	    case MLPOS:
		if (((LPMULTIWINPOS)lpTemp)->lpWinPosInfo)
		    WinFree((LPSTR)
			(((LPMULTIWINPOS)lpTemp)->lpWinPosInfo));
		break;

	    case MLHANDLE:
		TWIN_DeleteHandle((LPHANDLEINFO)lpTemp);
		break;

	    case MLLBOX:
		//TWIN_DeleteLBox((LPLISTBOXINFO) lpTemp);
		//break;
		return 0;

	    case MLFONT:
	    case MLREGION:
	    case MLPEN:
	    case MLBRUSH:
	    case MLPALETTE:
		TWIN_DeleteObject(lpTemp);
		break;
	    case MLBITMAP:
		//TWIN_DeleteObject(lpTemp);
		//break;
		return 0;

	    case MLDC:
		TWIN_DeleteDC((HDC32) lpTemp);
		break;

	    case MLICON:
		//TWIN_DeleteIcon((LPTWIN_ICONINFO) lpTemp);
		//break;
		return 0;

	    case MLCURSOR:
		//TWIN_DeleteCursor((LPCURSORINFO) lpTemp);
		//break;
		return 0;

	    case MLMODULE:
	    case MLTASK:
	    case MLSBAR:
	    case MLMETAFILE:
	    default:
		printf("no delete handler %x %s\n",
			wSignature,
			sigof(1,hObj,lpTemp->wObjSignature));
		debuggerbreak();
		break;
	}
	//logstr(lfobj,"OBJMGR %x S=%s #=%d F=%x\n",
	//	hObj,
	//	sigstr(lpTemp->wObjSignature),
	//	lpTemp->wRefCount,
	//	lpTemp->wFlags);

	freehandle(hObj);

	return 0;
}

LPOBJHEAD
HandleObj(int Func,WORD wSignature,LPVOID lpv)
{
    LPOBJHEAD lpTemp;
    int Size = 0;
    WORD Type, Category;
    HANDLE hObj = 0;

    Type = LOBYTE(wSignature);
    Category = HIBYTE(wSignature);

    switch(Func) {
	/* Get new object handle and pointer */
	case HM_GETHND:		
	    /* start off with NO handle */
	    *((LPHANDLE)lpv) = 0;

	    /* must have category and type */
	    if (!Category || !Type)
		return (LPOBJHEAD)GetHandleErr(2,0,wSignature,0);

	    /* how big is the object */
	    switch (wSignature) {
		case MLDC:
		    Size = sizeof(DCINFO);
		    break;
		case MLICON:
		    Size = sizeof(TWIN_ICONINFO);
		    break;
		case MLBITMAP:
		    Size = sizeof(TWIN_IMAGEINFO);
		    break;
		case MLHANDLE:
		case MLREGION:
		    Size = sizeof(HANDLEINFO);
		    break;
		case MLMODULE:
		    Size = sizeof(MODULEINFO);
		    break;
		case MLTASK:
		    Size = sizeof(TASKINFO);
		    break;
		case MLBRUSH:
		    Size = sizeof(BRUSHINFO);
		    break;
		case MLPEN:
		    Size = sizeof(PENINFO);
		    break;
		case MLFONT:
		    Size = sizeof(FONTINFO);
		    break;
		case MLPALETTE:
		    Size = sizeof(PALETTEINFO);
		    break;
		case MLWIN:
		    Size = sizeof(WININFO);
		    break;
		case MLPOS:
		    Size = sizeof(MULTIWINPOS);
		    break;
		case MLLBOX:
		    Size = sizeof(LISTBOXINFO);
		    break;
		case MLSBAR:
		    Size = sizeof(TWINSCROLLINFO);
		    break;
		case MLMETAFILE:
		    Size = sizeof(METAFILEINFO);
		    break;
		case MLCURSOR:
		    Size = sizeof(CURSORINFO);
		    break;
		default:
		    return (LPOBJHEAD)GetHandleErr(3,0,wSignature,0);
	    }
	    
	    /* get a new handle, and allocate memory for object */
	    lpTemp = newhandle(Size);

	    /* set its signature */
	    lpTemp->wObjSignature = MAKESIGNATURE(Category,Type);

	    /* return the handle we retrieved */
	    *((LPHANDLE)lpv) = lpTemp->hObj;

#ifdef DEBUG_OBJECTS
	    logstr(lfobj,"OBJMGR %x f=%s sig=%s refcnt=%d flag=%x\n",
	    	lpTemp->hObj,cmdnames[Func],
	     	sigof(Func,lpTemp->hObj,lpTemp->wObjSignature),
	     	lpTemp->wRefCount,
	     	lpTemp->wFlags);
#endif

	    return ((LPOBJHEAD)lpTemp);

	/* Check handle validity */
	case HM_CHECKHND:	{

	    /* get the HANDLE to dereference */
	    if (!((DWORD)(hObj = (HANDLE)(DWORD)lpv)))
	    	return (LPOBJHEAD)FALSE;

	    /* get its pointer */
	    if (!(lpTemp = getptr(hObj,0)))
			return (LPOBJHEAD)FALSE;

	    /* validate the object */
	    if (Category && (HIBYTE(lpTemp->wObjSignature) != Category))
		return (LPOBJHEAD)FALSE;

	    if (Type && (LOBYTE(lpTemp->wObjSignature) != Type))
		return (LPOBJHEAD)FALSE;

	    /* bump its refcnt */
	    lpTemp->wRefCount++;

#ifdef DEBUG_OBJECTS
	    logstr(LF_SYS,"OBJENGINE %x f=%s sig=%s refcnt=%d flag=%x\n",
	    	hObj,cmdnames[Func],
	     	sigof(Func,hObj,lpTemp->wObjSignature),
	     	lpTemp->wRefCount,
	     	lpTemp->wFlags);
#endif

	    return (LPOBJHEAD)lpTemp;
	}

	/* Return object ptr with a given handle */
	case HM_GETOBJ:		

	    if (!((DWORD)(hObj = (HANDLE)(DWORD)lpv) ))
	    	return (LPOBJHEAD)GetHandleErr(4,hObj,wSignature,0);

	    lpTemp = getptr(hObj,1);

	    if (lpTemp) {

		    if (Category && (HIBYTE(lpTemp->wObjSignature) != Category))
			return (LPOBJHEAD)GetHandleErr(8,hObj,wSignature,Category);

		    if (Type && (LOBYTE(lpTemp->wObjSignature) != Type))
			return (LPOBJHEAD)GetHandleErr(9,hObj,wSignature,Type);

		    lpTemp->wRefCount++;

#ifdef DEBUG_OBJECTS
		    logstr(LF_SYS,"OBJENGINE %x f=%s sig=%s #=%d flag=%x\n",
		    	hObj,cmdnames[Func],
		    	sigof(Func,hObj,lpTemp->wObjSignature),
		    	lpTemp->wRefCount,
		    	lpTemp->wFlags);
#endif
	    }
	    return lpTemp;

	case HM_LOCKHND:
	case HM_UNLOCKHND: 
	    if (!((DWORD)(hObj = (HANDLE)(DWORD)lpv) ))
	    	return (LPOBJHEAD)GetHandleErr(4,hObj,wSignature,0);

	    lpTemp = getptr(hObj,1);
	    if (lpTemp) {
		int nCount = (int)(short)((LPOBJHEAD)lpTemp)->wRefCount;

		switch(Func) {
		case HM_LOCKHND:
	    		lpTemp->wRefCount++;
			break;
	    	case HM_UNLOCKHND:
			if (nCount > 0)
	    			lpTemp->wRefCount--;

			if((lpTemp->wRefCount == 0) && (lpTemp->wFlags & OE_DELETE)) {
				/* don't call through this again */
				lpTemp->wRefCount = 1;
				deleteobject(hObj,lpTemp,wSignature);
			}
			break;
	    	}

	    	//logstr(LF_SYS,"OBJENGINE %x f=%s sig=%s refcnt=%d flag=%x\n", 
	    	//	lpTemp->wRefCount,
	    	//	lpTemp->wFlags);
			
	    }	
	    return (LPOBJHEAD)lpTemp;

	/* add object ptr to the free list */
	case HM_FREEHND:
	    if (!((DWORD)(hObj = (HANDLE)(DWORD)lpv) ))
	    	return (LPOBJHEAD)GetHandleErr(4,hObj,wSignature,0);

	    lpTemp = getptr(hObj,1);

	    if (lpTemp) {
		deleteobject(hObj,lpTemp,wSignature);
	    }

	    return (LPOBJHEAD)(-1);

	/* Dump handle table to log file */
	case HM_DUMPTABLE:

		DEBUG_DumpHandle();
		return 0;

	default:
	    debuggerbreak();
	    break;
    }
    return (LPOBJHEAD)GetHandleErr(1,(HANDLE)Func,wSignature,0);
}

void
DEBUG_DumpHandle()
{
	int IndexHi,IndexLo;
	LPOBJHEAD lpObj;
	char *lpString;
	int	n;

	logstr(lfobj,"OBJMGR: DumpHandle Table\n");
	if(lplpArray == 0)
		return;

	n = 0;
	for (IndexHi = 0; IndexHi < 256 ; IndexHi++)

		for (IndexLo = 0; IndexLo < 256; IndexLo++) {

		    if(lplpArray[IndexHi] == 0)
			continue;

	      	    lpObj = (LPOBJHEAD)lplpArray[IndexHi][IndexLo];

		    if(lpObj == 0)
			continue;
	
		    lpString = sigstr(lpObj->wObjSignature);

	      	    LOGSTR((lfobj,"%.4d %4.4x S=%4.4x %s %s F=%x %s%s#=%d *=%p\n",
			n++,
			lpObj->hObj,
			lpObj->wObjSignature,
			sigof(-1,lpObj->hObj,lpObj->wObjSignature),
			lpString,
			lpObj->wFlags,
			lpObj->wFlags&1?"LOCKED ":"",
			lpObj->wFlags&2?"DELETED ":"",
			lpObj->wRefCount, 
			lpObj));
	  }

	  LOGSTR((lfobj,"OBJMGR: End of dump.\n"));
	  return;
}

/*------------------------- < Full Revision History > ----------------------
** $Log: ObjEngine.c,v $
** Revision 1.9  1999/12/14 17:49:17  robf
** code cleanup
**
** Revision 1.8  1999/12/08 07:56:06  robf
** added flag to fix refcnt
**
** Revision 1.6  1999/07/09 21:55:43  mwalsh
** Updated comments to standard header and footer
**
**-------------------------------------------------------------------------*/
